
RELEASE_VERSION='0.0.362525997'
BUILD_DATE='Tue Aug 31 13:06:14 UTC 2021'
BUILD_MACHINE_INFO='Linux runner-mvj-ljo-project-28359933-concurrent-0qbzxl 5.4.129-62.227.amzn2.x86_64 #1 SMP Wed Jul 7 00:08:43 UTC 2021 x86_64 GNU/Linux'
GITLAB_CI_JOB_ID='1549299892'
GIT_COMMIT_SHA='68c406d96169dbbadbf203087f9439885b2d6324'
GIT_BRANCH='main'
